@@ -196,6 +196,8 @@ permission_error(output, binary_stream, Stream) -->
     [ 'No permission to write characters to binary stream `~p'''-[Stream] ].
 permission_error(open, source_sink, alias(Alias)) -->
     [ 'No permission to reuse alias "~p": already taken'-[Alias] ].
+permission_error(tnot, non_tabled_procedure, Pred) -->
+    [ 'The argument of tnot/1 is not tabled: ~p'-[Pred] ].
 permission_error(Action, Type, Object) -->
     [ 'No permission to ~w ~w `~p'''-[Action, Type, Object] ].
 
