@@ -943,9 +943,10 @@ define_or_generate(Pred) :-
     '$get_predicate_attribute'(Pred, quasi_quotation_syntax, 1).
 '$predicate_property'(defined, Pred) :-
     '$get_predicate_attribute'(Pred, defined, 1).
-'$predicate_property'(tabled(default), M:Pred) :-
-    '$c_current_predicate'(_, M:'$tabled'(_)),
-    M:'$tabled'(Pred).
+'$predicate_property'(tabled(Mode), Pred) :-
+    '$get_predicate_attribute'(Pred, tabled, 1),
+    '$tbl_implementation'(Pred, M:Head),
+    M:'$tabled'(Head, Mode).
 
 system_undefined(user:prolog_trace_interception/4).
 system_undefined(user:prolog_exception_hook/4).
@@ -1186,7 +1187,11 @@ trie_property(node_count(_)).
 trie_property(value_count(_)).
 trie_property(size(_)).
 trie_property(hashed(_)).
-
+                                                % below only when -DO_TRIE_STATS
+trie_property(lookup_count(_)).                 % is enabled in pl-trie.h
+trie_property(gen_call_count(_)).
+trie_property(gen_exit_count(_)).
+trie_property(gen_fail_count(_)).
 
 
                 /********************************
