@@ -35,23 +35,28 @@
 */
 
 :- module('$tabling',
-          [ (table)/1,                  % +PI ...
+          [ (table)/1,                  % :PI ...
+            untable/1,                  % :PI ...
 
-            (tnot)/1,                     % :Goal
+            (tnot)/1,                   % :Goal
+            undefined/0,
 
             current_table/2,            % :Variant, ?Table
             abolish_all_tables/0,
             abolish_table_subgoals/1,   % :Subgoal
 
             start_tabling/2,            % +Wrapper, :Worker
+            start_subsumptive_tabling/2,% +Wrapper, :Worker
             start_tabling/4,            % +Wrapper, :Worker, :Variant, ?ModeArgs
 
-            '$wrap_tabled'/1,		% :Head
+            '$wrap_tabled'/2,		% :Head, +Mode
             '$moded_wrap_tabled'/4,	% :Head, +ModeTest, +Variant, +Moded
             '$wfs_call'/2               % :Goal, -Delays
           ]).
 
 :- meta_predicate
+    table(:),
+    untable(:),
     tnot(0),
     start_tabling(+, 0),
     start_tabling(+, 0, +, ?),
@@ -78,7 +83,13 @@ goal_expansion(tdebug(Topic, Fmt, Args), Expansion) :-
     ).
 goal_expansion(tdebug(Goal), Expansion) :-
     (   current_prolog_flag(prolog_debug, true)
-    ->  Expansion = (Goal->true;print_message(error, goal_failed(Goal)))
+    ->  Expansion = (   debugging(tabling(_))
+                    ->  (   Goal
+                        ->  true
+                        ;   print_message(error, goal_failed(Goal))
+                        )
+                    ;   true
+                    )
     ;   Expansion = true
     ).
 
@@ -105,11 +116,13 @@ user_goal(Goal, UGoal) :-
 
 :- endif.
 
-%!  table(+PredicateIndicators)
+%!  table(:PredicateIndicators)
 %
-%   Prepare the given PredicateIndicators for   tabling. Can only be
-%   used as a directive. The example   below  prepares the predicate
-%   edge/2 and the non-terminal statement//1 for tabled execution.
+%   Prepare the given PredicateIndicators for tabling. This predicate is
+%   normally used as a directive,  but   SWI-Prolog  also allows runtime
+%   conversion of non-tabled predicates to  tabled predicates by calling
+%   table/1. The example below prepares  the   predicate  edge/2 and the
+%   non-terminal statement//1 for tabled execution.
 %
 %     ==
 %     :- table edge/2, statement//1.
@@ -126,8 +139,91 @@ user_goal(Goal, UGoal) :-
 %   _Mode directed tabling_ is  discussed   in  the general introduction
 %   section about tabling.
 
-table(PIList) :-
-    throw(error(context_error(nodirective, table(PIList)), _)).
+table(M:PIList) :-
+    setup_call_cleanup(
+        '$set_source_module'(OldModule, M),
+        expand_term((:- table(PIList)), Clauses),
+        '$set_source_module'(OldModule)),
+    dyn_tabling_list(Clauses, M).
+
+dyn_tabling_list([], _).
+dyn_tabling_list([H|T], M) :-
+    dyn_tabling(H, M),
+    dyn_tabling_list(T, M).
+
+dyn_tabling((:- multifile(PI)), M) :-
+    !,
+    multifile(M:PI),
+    dynamic(M:PI).
+dyn_tabling(:- initialization(Wrap, now), M) :-
+    !,
+    M:Wrap.
+dyn_tabling('$tabled'(Head, TMode), M) :-
+    (   clause(M:'$tabled'(Head, OMode), true, Ref),
+        (   OMode \== TMode
+        ->  erase(Ref),
+            fail
+        ;   true
+        )
+    ->  true
+    ;   assertz(M:'$tabled'(Head, TMode))
+    ).
+dyn_tabling('$table_mode'(Head, Variant, Moded), M) :-
+    (   clause(M:'$table_mode'(Head, Variant0, Moded0), true, Ref)
+    ->  (   t(Head, Variant, Moded) =@= t(Head, Variant0, Moded0)
+        ->  true
+        ;   erase(Ref),
+            assertz(M:'$table_mode'(Head, Variant, Moded))
+        )
+    ;   assertz(M:'$table_mode'(Head, Variant, Moded))
+    ).
+
+%!  untable(M:PIList) is det.
+%
+%   Remove tabling for the predicates in  PIList.   This  can be used to
+%   undo the effect of table/1 at runtime.   In addition to removing the
+%   tabling instrumentation this also removes possibly associated tables
+%   using abolish_table_subgoals/1.
+%
+%   @arg PIList is a comma-list that is compatible ith table/1.
+
+untable(M:PIList) :-
+    untable(PIList, M).
+
+untable(Var, _) :-
+    var(Var),
+    !,
+    '$instantiation_error'(Var).
+untable((A,B), M) :-
+    !,
+    untable(A, M),
+    untable(B, M).
+untable(Name//Arity, M) :-
+    atom(Name), integer(Arity), Arity >= 0,
+    !,
+    Arity1 is Arity+2,
+    untable(Name/Arity1, M).
+untable(Name/Arity, M) :-
+    !,
+    functor(Head, Name, Arity),
+    (   predicate_property(M:Head, tabled(_))
+    ->  abolish_table_subgoals(M:Head),
+        dynamic(M:'$tabled'/2),
+        dynamic(M:'$table_mode'/3),
+        retractall(M:'$tabled'(Head, _TMode)),
+        retractall(M:'$table_mode'(Head, _Variant, _Moded)),
+        unwrap_predicate(M:Name/Arity, table),
+        '$set_predicate_attribute'(M:Head, tabled, false)
+    ;   true
+    ).
+untable(Head, M) :-
+    callable(Head),
+    !,
+    functor(Head, Name, Arity),
+    untable(Name/Arity, M).
+untable(TableSpec, _) :-
+    '$type_error'(table_desclaration, TableSpec).
+
 
 %!  start_tabling(:Wrapper, :Implementation)
 %
@@ -139,9 +235,16 @@ table(PIList) :-
 %   @compat This interface may change or disappear without notice
 %           from future versions.
 
-'$wrap_tabled'(Head) :-
+'$wrap_tabled'(Head, variant) :-
+    !,
+    '$set_predicate_attribute'(Head, tabled, true),
     '$wrap_predicate'(Head, table, Wrapped,
                       start_tabling(Head, Wrapped)).
+'$wrap_tabled'(Head, subsumptive) :-
+    !,
+    '$set_predicate_attribute'(Head, tabled, true),
+    '$wrap_predicate'(Head, table, Wrapped,
+                      start_subsumptive_tabling(Head, Wrapped)).
 
 start_tabling(Wrapper, Worker) :-
     '$tbl_variant_table'(Wrapper, Trie, Status, Skeleton),
@@ -162,6 +265,33 @@ start_tabling(Wrapper, Worker) :-
         shift(call_info(Skeleton, Status))
     ).
 
+
+%!  start_subsumptive_tabling(:Wrapper, :Implementation)
+
+start_subsumptive_tabling(Wrapper, Worker) :-
+    (   '$tbl_existing_variant_table'(Wrapper, Trie, Status, Skeleton)
+    ->  (   Status == complete
+        ->  '$tbl_answer_update_dl'(Trie, Skeleton)
+        ;    shift(call_info(Skeleton, Status))
+        )
+    ;   more_general_table(Wrapper, ATrie),
+        '$tbl_table_status'(ATrie, complete, Wrapper, Skeleton)
+    ->  '$tbl_answer_update_dl'(ATrie, Skeleton)
+    ;   '$tbl_variant_table'(Wrapper, Trie, _0Status, Skeleton),
+        tdebug(_0Status == fresh),
+        '$tbl_create_subcomponent'(SCC, Trie),
+        tdebug(user_goal(Wrapper, Goal)),
+        tdebug(schedule, 'Created component ~d for ~p', [SCC, Goal]),
+        setup_call_catcher_cleanup(
+            true,
+            run_leader(Skeleton, Worker, Trie, SCC, LStatus),
+            Catcher,
+            finished_leader(Catcher, SCC, Wrapper)),
+        tdebug(schedule, 'Leader ~p done, status = ~p', [Goal, LStatus]),
+        done_leader(LStatus, SCC, Skeleton, Trie)
+    ).
+
+
 done_leader(complete, _SCC, Skeleton, Trie) :-
     !,
     '$tbl_answer_update_dl'(Trie, Skeleton).
@@ -238,8 +368,9 @@ delim(Wrapper, Worker, WorkList, Delays) :-
         !
     ;   SourceCall = call_info(SrcSkeleton, SourceWL),
         '$tbl_add_global_delays'(Delays, AllDelays),
-        tdebug(wl_goal(SourceWL, SrcWrapper, _)),
-        tdebug(schedule, 'Suspended ~p, for solving ~p', [SrcWrapper, Wrapper]),
+        tdebug(wl_goal(SourceWL, SrcGoal, _)),
+        tdebug(wl_goal(WorkList, DstGoal, _)),
+        tdebug(schedule, 'Suspended ~p, for solving ~p', [SrcGoal, DstGoal]),
         '$tbl_wkl_add_suspension'(
             SourceWL,
             dependency(SrcSkeleton, Continuation, Wrapper, WorkList, AllDelays))
@@ -251,6 +382,7 @@ delim(Wrapper, Worker, WorkList, Delays) :-
 %   answer trie in the Variant and ModeArgs.
 
 '$moded_wrap_tabled'(Head, ModeTest, WrapperNoModes, ModeArgs) :-
+    '$set_predicate_attribute'(Head, tabled, true),
     '$wrap_predicate'(Head, table, Wrapped,
                       (   ModeTest,
                           start_tabling(Head, Wrapped, WrapperNoModes, ModeArgs)
@@ -394,7 +526,8 @@ completion_step(WorkList) :-
         tdebug(wl_goal(WorkList, SourceGoal, _)),
         tdebug(wl_goal(TargetWorklist, TargetGoal, _Skeleton)),
         (   ModeArgs == Reserved
-        ->  tdebug(delay_goals(Delays, Cond)),
+        ->  tdebug('$tbl_add_global_delays'(Delays, AllDelays)),
+            tdebug(delay_goals(AllDelays, Cond)),
             tdebug(schedule, 'Resuming ~p, calling ~p with ~p (delays = ~p)',
                    [TargetGoal, SourceGoal, Answer, Cond]),
             Goal = Answer,
@@ -416,10 +549,9 @@ completion_step(WorkList) :-
 %!  tnot(:Goal)
 %
 %   Tabled negation.
-%
-%   @tbd: verify Goal is actually tabled.
 
-tnot(Goal) :-
+tnot(Goal0) :-
+    '$tbl_implementation'(Goal0, Goal),         % verifies Goal is tabled
     '$tbl_variant_table'(Goal, Trie, Status, Skeleton),
     (   '$tbl_answer_dl'(Trie, _, true)
     ->  fail
@@ -483,47 +615,42 @@ reset_delays :-
     '$tbl_delay_list'(DL0),
     reset_delays,
     call(Goal),
-    delay_list(M, Delays),
-    '$append'(DL0, Delays, DL),
+    '$tbl_delay_list'(DL1),
+    delay_goals(DL1, M, Delays),
+    '$append'(DL0, DL1, DL),
     '$tbl_set_delay_list'(DL).
 
-delay_list(M, Delays) :-
-    '$tbl_delay_list'(DL),
-    delay_goals(DL, M, Delays).
-
 delay_goals([], _, true) :-
     !.
 delay_goals([AT+AN|T], M, Goal) :-
     !,
     (   integer(AN)
-    ->  at_delay_goal(AT, G0, Answer),
+    ->  at_delay_goal(AT, M, G0, Answer),
         trie_term(AN, Answer)
-    ;   AN = G0
+    ;   '$tbl_table_status'(AT, _Status, G0, AN)
     ),
-    unqualify_goal(G0, M, G1),
-    GN = G1,
+    GN = G0,
     (   T == []
     ->  Goal = GN
     ;   Goal = (GN,GT),
         delay_goals(T, M, GT)
     ).
 delay_goals([AT|T], M, Goal) :-
-    at_delay_goal(AT, G0, _Skeleton),
-    unqualify_goal(G0, M, G1),
-    GN = tnot(G1),
+    at_delay_goal(AT, M, G0, _Skeleton),
+    GN = tnot(G0),
     (   T == []
     ->  Goal = GN
     ;   Goal = (GN,GT),
         delay_goals(T, M, GT)
     ).
 
-at_delay_goal(tnot(Trie), tnot(Goal), Skeleton) :-
+at_delay_goal(tnot(Trie), M, tnot(Goal), Skeleton) :-
     !,
     '$tbl_table_status'(Trie, _Status, Wrapper, Skeleton),
-    unqualify_goal(Wrapper, user, Goal).
-at_delay_goal(Trie, Goal, Skeleton) :-
+    unqualify_goal(Wrapper, M, Goal).
+at_delay_goal(Trie, M, Goal, Skeleton) :-
     '$tbl_table_status'(Trie, _Status, Wrapper, Skeleton),
-    unqualify_goal(Wrapper, user, Goal).
+    unqualify_goal(Wrapper, M, Goal).
 
 unqualify_goal(M:Goal, M, Goal0) :-
     !,
@@ -553,9 +680,11 @@ abolish_all_tables :-
 
 abolish_table_subgoals(M:SubGoal) :-
     '$tbl_variant_table'(VariantTrie),
+    !,
     current_module(M),
     forall(trie_gen(VariantTrie, M:SubGoal, Trie),
            '$tbl_destroy_table'(Trie)).
+abolish_table_subgoals(_).
 
 
                  /*******************************
@@ -584,34 +713,43 @@ current_table(M:Variant, Trie) :-
 :- dynamic
     system:term_expansion/2.
 
-wrappers(Var) -->
+wrappers(Spec) -->
+    wrappers(Spec, #{}).
+
+wrappers(Var, _) -->
     { var(Var),
       !,
       '$instantiation_error'(Var)
     }.
-wrappers((A,B)) -->
+wrappers(Spec as Options, Opts0) -->
     !,
-    wrappers(A),
-    wrappers(B).
-wrappers(Name//Arity) -->
+    { table_options(Options, Opts0, Opts) },
+    wrappers(Spec, Opts).
+wrappers((A,B), Opts) -->
+    !,
+    wrappers(A, Opts),
+    wrappers(B, Opts).
+wrappers(Name//Arity, Opts) -->
     { atom(Name), integer(Arity), Arity >= 0,
       !,
       Arity1 is Arity+2
     },
-    wrappers(Name/Arity1).
-wrappers(Name/Arity) -->
-    { atom(Name), integer(Arity), Arity >= 0,
+    wrappers(Name/Arity1, Opts).
+wrappers(Name/Arity, Opts) -->
+    { '$option'(mode(TMode), Opts, variant),
+      atom(Name), integer(Arity), Arity >= 0,
       !,
       functor(Head, Name, Arity),
       prolog_load_context(module, Module),
       '$tbl_trienode'(Reserved)
     },
-    [ '$tabled'(Head),
+    [ '$tabled'(Head, TMode),
       '$table_mode'(Head, Head, Reserved),
-      (:- initialization('$wrap_tabled'(Module:Head), now))
+      (:- initialization('$wrap_tabled'(Module:Head, TMode), now))
     ].
-wrappers(ModeDirectedSpec) -->
-    { callable(ModeDirectedSpec),
+wrappers(ModeDirectedSpec, Opts) -->
+    { '$option'(mode(TMode), Opts, variant),
+      callable(ModeDirectedSpec),
       !,
       functor(ModeDirectedSpec, Name, Arity),
       functor(Head, Name, Arity),
@@ -620,20 +758,36 @@ wrappers(ModeDirectedSpec) -->
       prolog_load_context(module, Module),
       mode_check(Moded, ModeTest),
       (   ModeTest == true
-      ->  WrapClause = '$wrap_tabled'(Module:Head)
+      ->  WrapClause = '$wrap_tabled'(Module:Head, TMode)
       ;   WrapClause = '$moded_wrap_tabled'(Module:Head, ModeTest,
           Module:Variant, Moded)
       )
     },
-    [ '$tabled'(Head),
+    [ '$tabled'(Head, TMode),
       '$table_mode'(Head, Variant, Moded),
       (:- initialization(WrapClause, now))
     | UpdateClauses
     ].
-wrappers(TableSpec) -->
+wrappers(TableSpec, _Opts) -->
     { '$type_error'(table_desclaration, TableSpec)
     }.
 
+%!  table_options(+Options, +OptDictIn, -OptDictOut)
+%
+%   Handler the ... as _options_ ... construct.
+
+table_options(Options, _Opts0, _Opts) :-
+    var(Options),
+    '$instantiation_error'(Options).
+table_options((A,B), Opts0, Opts) :-
+    table_options(A, Opts0, Opts1),
+    table_options(B, Opts1, Opts).
+table_options(subsumptive, Opts0, Opts1) :-
+    put_dict(mode, Opts0, subsumptive, Opts1).
+table_options(variant, Opts0, Opts1) :-
+    put_dict(mode, Opts0, variant, Opts1).
+
+
 %!  mode_check(+Moded, -TestCode)
 %
 %   Enforce the output arguments of a  mode-directed tabled predicate to
@@ -818,7 +972,7 @@ sum(S0, S1, S) :- S is S0+S1.
 		 *******************************/
 
 system:term_expansion((:- table(Preds)),
-                      [ (:- multifile('$tabled'/1)),
+                      [ (:- multifile('$tabled'/2)),
                         (:- multifile('$table_mode'/3)),
                         (:- multifile('$table_update'/4))
                       | Clauses
@@ -892,7 +1046,7 @@ delete_answers_for_failing_calls(Propagated) :-
 
 mark_succeeding_calls_as_answer_completed :-
     (   subgoal_residual_trie(ASGF, _ESGF),
-        (   '$tbl_answer_dl'(ASGF, _0Return, true)
+        (   '$tbl_answer_dl'(ASGF, _0Return, _True)
         ->  tdebug(trie_goal(ASGF, Answer, _0Return)),
             tdebug(trie_goal(ASGF, Goal, _0Return)),
             tdebug(ac(prune), '  Completed ~p on ~p', [Goal, Answer]),
@@ -962,8 +1116,22 @@ all_vars([H|T]) :-
 %   Derive answers for the variant represented   by  AnswerTrie based on
 %   the residual goals only.
 
+eval_subgoal_in_residual(AnswerTrie, _Return) :-
+    '$tbl_is_answer_completed'(AnswerTrie),
+    !,
+    undefined.
 eval_subgoal_in_residual(AnswerTrie, Return) :-
     '$tbl_answer'(AnswerTrie, Return, Condition),
     tdebug(trie_goal(AnswerTrie, Goal, Return)),
     tdebug(ac, 'Condition for ~p is ~p', [Goal, Condition]),
     eval_dl_in_residual(Condition).
+
+%!  undefined
+%
+%   Expresses the value _bottom_ from the well founded semantics.
+
+:- table
+    undefined/0.
+
+undefined :-
+    tnot(undefined).
