@@ -171,6 +171,7 @@ const debug_topic debug_topics[] =
 
   DEBUG_TOPIC(MSG_TRIE_PUT_TERM),
   DEBUG_TOPIC(MSG_TRIE_GC),
+  DEBUG_TOPIC(MSG_TRIE_GEN),
 
   DEBUG_TOPIC(MSG_THREAD_LOCAL),
 
@@ -186,6 +187,7 @@ const debug_topic debug_topics[] =
   DEBUG_TOPIC(MSG_TABLING_DELAY),
   DEBUG_TOPIC(MSG_TABLING_DELAY_VAR),
   DEBUG_TOPIC(MSG_TABLING_AC),
+  DEBUG_TOPIC(MSG_TABLING_MERGE),
 
   DEBUG_TOPIC(TABLING_NO_SIMPLIFY),
   DEBUG_TOPIC(TABLING_NO_AC),
