@@ -171,6 +171,7 @@
 
 #define MSG_TRIE_PUT_TERM	 200
 #define MSG_TRIE_GC		 201
+#define MSG_TRIE_GEN		 202
 
 #define MSG_THREAD_LOCAL	 240
 
@@ -186,6 +187,7 @@
 #define MSG_TABLING_DELAY	 304
 #define MSG_TABLING_DELAY_VAR	 305
 #define MSG_TABLING_AC		 306
+#define MSG_TABLING_MERGE	 307
 
 #define TABLING_NO_SIMPLIFY	 310		/* these affect behaviour */
 #define TABLING_NO_AC		 311
