@@ -437,7 +437,10 @@ unlinkSourceFileModule(SourceFile sf, Module m)
 
     if ( lookupHTable(m->procedures, (void*)def->functor->functor) ||
 	 PROCEDURE_dinit_goal->definition == def )	/* see (*) */
-    { if ( prev )
+    { if ( sf->current_procedure == proc )
+	sf->current_procedure = NULL;
+
+      if ( prev )
 	prev->next = cell->next;
       else
 	sf->procedures = cell->next;
@@ -791,6 +794,8 @@ startReconsultFile(SourceFile sf)
 	    continue;
 	  if ( true(def, P_MULTIFILE) && cl->owner_no != sf->index )
 	    continue;
+	  if ( true(def, P_DYNAMIC) && cl->owner_no == 0 )
+	    continue;
 
 	  cl->generation.erased = r->reload_gen;
 	}
@@ -1082,7 +1087,7 @@ associateSource(SourceFile sf, Procedure proc)
 }
 
 
-#define P_ATEND	(P_VOLATILE|P_PUBLIC|P_ISO|P_NOPROFILE|P_NON_TERMINAL)
+#define P_ATEND	(P_VOLATILE|P_PUBLIC|P_ISO|P_NOPROFILE|P_NON_TERMINAL|P_TABLED)
 
 int
 setAttrProcedureSource(SourceFile sf, Procedure proc,
