@@ -37,14 +37,29 @@
 #include "pl-tabling.h"
 #include "pl-copyterm.h"
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+We provide two answer completion strategies:
+
+  - Eager AC: #define O_AC_EAGER 1
+    Complete each component fully before continuing to the next.
+  - Lazy AC:  #undef O_AC_EAGER
+    Only complete the leader of a component.  This is what XSB is
+    doing.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#define O_AC_EAGER 1
+// #undef O_AC_EAGER
+
 #define record_t fastheap_term *
 #define PL_record(t)      term_to_fastheap(t PASS_LD)
 #define PL_recorded(r, t) put_fastheap(r, t PASS_LD)
 #define PL_erase(r)	  free_fastheap(r)
 
 static void	free_worklist(worklist *wl);
+static void	clean_worklist(worklist *wl);
 static void	free_worklist_set(worklist_set *wls, int freewl);
 static void	add_global_worklist(worklist *wl);
+static int	wl_has_work(const worklist *wl);
 static cluster *new_answer_cluster(worklist *wl, trie_node *first);
 static void	wkl_append_left(worklist *wl, cluster *c);
 static int	wkl_add_answer(worklist *wl, trie_node *node ARG_LD);
@@ -107,10 +122,14 @@ free_component(tbl_component *c, int flags)
     del_child_component(c->parent, c);
   if ( c->worklist )
     free_worklist_set(c->worklist, WLFS_FREE_NONE);
+  if ( c->delay_worklists )
+    free_worklist_set(c->delay_worklists, WLFS_FREE_NONE);
   if ( c->created_worklists )
     free_worklist_set(c->created_worklists, WLFS_FREE_ALL);
   if ( c->children )
     free_components_set(c->children, flags|FC_CHILD);
+  if ( c->merged )
+    free_components_set(c->merged, flags|FC_CHILD);
 
   PL_free(c);
 }
@@ -129,21 +148,6 @@ add_child_component(tbl_component *parent, tbl_component *child)
   addBuffer(&cs->members, child, tbl_component*);
 }
 
-/*
-static void
-print_component_set(char *why, component_set *cs)
-{ tbl_component **bp = baseBuffer(&cs->members, tbl_component*);
-  tbl_component **tp = topBuffer(&cs->members, tbl_component*);
-
-  Sdprintf("%s:", why);
-  for(;	bp < tp; bp++)
-  { tbl_component *c = *bp;
-    Sdprintf(" %p", c);
-  }
-  Sdprintf("\n");
-}
-*/
-
 static void
 del_child_component(tbl_component *parent, tbl_component *child)
 { component_set *cs = parent->children;
@@ -176,17 +180,30 @@ Merge all subcomponets of c into c.   The properties of the subcomponets
 are destroyed and .status is set to SCC_MERGED.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static void merge_children(tbl_component *c, component_set *cs);
+static void merge_children(tbl_component *c, tbl_component *m);
 static void merge_one_component(tbl_component *c, tbl_component *m);
-static void wls_set_component(worklist_set *wls, tbl_component *c);
+static void wls_set_component(worklist_set *wls, size_t size0, tbl_component *c);
+
+static size_t
+wls_size(const worklist_set *wls)
+{ return wls ? entriesBuffer(&wls->members, worklist*) : 0;
+}
+
 
 static void
 merge_component(tbl_component *c)
-{ if ( c->children )
-    merge_children(c, c->children);
+{ size_t s_global  = wls_size(c->worklist);
+  size_t s_created = wls_size(c->created_worklists);
 
-  wls_set_component(c->worklist, c);
-  wls_set_component(c->created_worklists, c);
+  if ( c->children )
+    merge_children(c, c);
+
+  wls_set_component(c->worklist, s_global, c);
+  wls_set_component(c->created_worklists, s_created, c);
+
+  DEBUG(MSG_TABLING_MERGE,
+	Sdprintf("Grown SCC %p from %zd to %zd worklists\n",
+		 c, s_created, wls_size(c->created_worklists)));
 }
 
 
@@ -194,10 +211,10 @@ static void
 wl_set_component(worklist *wl, tbl_component *c)
 { wl->component = c;
   wl->executing = FALSE;
-  if ( !wl->in_global_wl )
+  if ( !wl->in_global_wl && wl_has_work(wl) )
     add_global_worklist(wl);
   if ( wl->negative )
-  { DEBUG(MSG_TABLING_NEG,
+  { DEBUG(MSG_TABLING_MERGE,
 	  Sdprintf("Merging negative literal into SCC %zd\n",
 		   pointerToInt(c)));
     if ( c->neg_status == SCC_NEG_NONE )
@@ -209,21 +226,59 @@ wl_set_component(worklist *wl, tbl_component *c)
 
 
 static void
-wls_set_component(worklist_set *wls, tbl_component *c)
+wls_set_component(worklist_set *wls, size_t size0, tbl_component *c)
 { worklist **base = baseBuffer(&wls->members, worklist*);
   worklist **top  = topBuffer(&wls->members, worklist*);
 
+#ifdef O_DEBUG
+  size_t old = 0;
+  for(; base < top; base++, old++)
+  { if ( old < size0 )
+      assert((*base)->component == c);
+    else
+      wl_set_component(*base, c);
+  }
+#else
+  base += size0;			/* skip old ones */
   for(; base < top; base++)
     wl_set_component(*base, c);
+#endif
 }
 
 static void
-merge_children(tbl_component *c, component_set *cs)
-{ tbl_component **bp = baseBuffer(&cs->members, tbl_component*);
-  tbl_component **tp = topBuffer(&cs->members, tbl_component*);
+merge_component_sets(component_set **into, component_set **from)
+{ typedef tbl_component* Component;
 
-  for(;	bp < tp; bp++)
-    merge_one_component(c, *bp);
+  if ( *into && *from )
+  { tbl_component **s = baseBuffer(&(*from)->members, tbl_component*);
+    size_t        cnt = entriesBuffer(&(*from)->members, tbl_component*);
+    Buffer	    b = &(*into)->members;
+
+    addMultipleBuffer(b, s, cnt, Component);
+    free_components_set(*from, 0);
+    *from = NULL;
+  } else if ( *from )
+  { *into = *from;
+    *from = NULL;
+  }
+}
+
+
+/* Merge all components of cs into c */
+
+static void
+merge_children(tbl_component *c, tbl_component *m)
+{ component_set *cs;
+
+  if ( (cs=m->children) )
+  { tbl_component **bp = baseBuffer(&cs->members, tbl_component*);
+    tbl_component **tp = topBuffer(&cs->members, tbl_component*);
+
+    for( ; bp < tp; bp++)
+      merge_one_component(c, *bp);
+
+    merge_component_sets(&m->merged, &m->children);
+  }
 }
 
 
@@ -253,10 +308,9 @@ merge_one_component(tbl_component *c, tbl_component *m)
   if ( m->status != SCC_ACTIVE )
     return;
 
-  if ( m->children )
-    merge_children(c, m->children);
+  merge_children(c, m);
 
-  DEBUG(MSG_TABLING_WORK,
+  DEBUG(MSG_TABLING_MERGE,
 	Sdprintf("Merged %p into %p, %zd worklists, %zd created\n",
 		 m, c,
 		 entriesBuffer(&m->worklist->members, worklist*),
@@ -264,6 +318,8 @@ merge_one_component(tbl_component *c, tbl_component *m)
 
   merge_worklists(&c->worklist, &m->worklist);
   merge_worklists(&c->created_worklists, &m->created_worklists);
+  merge_worklists(&c->delay_worklists, &m->delay_worklists);
+
   m->status = SCC_MERGED;
 }
 
@@ -271,67 +327,56 @@ merge_one_component(tbl_component *c, tbl_component *m)
 		 *           WORKLISTS		*
 		 *******************************/
 
+static worklist_set *
+new_worklist_set(worklist *wl)
+{ worklist_set *wls = PL_malloc(sizeof(*wls));
+
+  initBuffer(&wls->members);
+  addBuffer(&wls->members, wl, worklist*);
+
+  return wls;
+}
+
+
 static void
 add_global_worklist(worklist *wl)
 { tbl_component *c = wl->component;
   worklist_set *wls;
 
   if ( !(wls=c->worklist) )
-  { wls	= c->worklist = PL_malloc(sizeof(*c->worklist));
-    initBuffer(&wls->members);
-  }
+    c->worklist = new_worklist_set(wl);
+  else
+    addBuffer(&wls->members, wl, worklist*);
 
-  addBuffer(&wls->members, wl, worklist*);
   wl->in_global_wl = TRUE;
 }
 
 
-#ifdef O_DEBUG
-static int
-is_in_global_worklist(worklist *wl)
+static void
+add_delay_worklist(worklist *wl)
 { tbl_component *c = wl->component;
   worklist_set *wls;
 
-  if ( (wls = c->worklist) )
-  { worklist **wlp = baseBuffer(&wls->members, worklist*);
-    worklist **elp = topBuffer(&wls->members, worklist*);
-
-    for( ; wlp < elp; wlp++)
-    { if ( *wlp == wl )
-	return TRUE;
-    }
-  }
-
-  return FALSE;
+  if ( !(wls=c->delay_worklists) )
+    c->delay_worklists = new_worklist_set(wl);
+  else
+    addBuffer(&wls->members, wl, worklist*);
 }
-#endif
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Normal completion is done. There  may   be  worklists that are suspended
 using negation_suspend/3. We wake  these  up   by  adding  a  new answer
 cluster with a NULL node.
-
-For WFS we need to pass the condition we represent in the delay list.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static worklist *
 negative_worklist(tbl_component *scc ARG_LD)
-{ worklist **wlp = baseBuffer(&scc->created_worklists->members, worklist*);
-  worklist **top = topBuffer(&scc->created_worklists->members, worklist*);
+{ if ( scc->delay_worklists )
+  { while( !isEmptyBuffer(&scc->delay_worklists->members) )
+    { worklist *wl = popBuffer(&scc->delay_worklists->members, worklist*);
 
-  if ( scc->neg_status == SCC_NEG_DELAY )
-  { DEBUG(MSG_TABLING_NEG,
-	  Sdprintf("Searching delayed negative literals in SCC %zd\n",
-		   pointerToInt(scc)));
-
-    wlp = baseBuffer(&scc->created_worklists->members, worklist*);
-    for(; wlp < top; wlp++)
-    { worklist *wl = *wlp;
-
-      if ( wl->negative &&
-	   !wl->neg_delayed &&
-	   !wl->has_answers )
+      if ( !wl->has_answers )	/* we have an unconditional answers, so no delay */
       { cluster *c;
 
 	wl->neg_delayed = TRUE;
@@ -370,15 +415,23 @@ negative_worklist(tbl_component *scc ARG_LD)
 }
 
 
+static int
+wl_has_work(const worklist *wl)
+{ return wl->riac && wl->riac->next;
+}
+
 static worklist *
 pop_worklist(tbl_component *c ARG_LD)
-{ worklist_set *wls;
+{ worklist_set *wls = c->worklist;
 
-  if ( (wls=c->worklist) && !isEmptyBuffer(&wls->members) )
-  { worklist *wl = popBuffer(&wls->members, worklist*);
-    wl->in_global_wl = FALSE;
+  if ( wls )
+  { while( !isEmptyBuffer(&wls->members) )
+    { worklist *wl = popBuffer(&wls->members, worklist*);
+      wl->in_global_wl = FALSE;
 
-    return wl;
+      if ( wl_has_work(wl) )
+	return wl;
+    }
   }
 
   return NULL;
@@ -410,12 +463,12 @@ add_newly_created_worklist(worklist *wl)
 }
 
 static void
-reset_newly_created_worklists(tbl_component *c)
+reset_newly_created_worklists(tbl_component *c, int flags)
 { worklist_set *wls;
 
   if ( c && (wls = c->created_worklists) )
   { c->created_worklists = NULL;
-    free_worklist_set(wls, WLFS_KEEP_COMPLETE);
+    free_worklist_set(wls, flags);
   }
 }
 
@@ -474,6 +527,21 @@ populate_answers(worklist *wl)
 		 *	 TABLE DELAY LISTS	*
 		 *******************************/
 
+static inline trie_node *
+REC_DELAY(record_t r)
+{ return (trie_node*)(((uintptr_t)r)|0x1);
+}
+
+static inline record_t
+UNREC_DELAY(trie_node *r)
+{ return (record_t)(((uintptr_t)r)&~(uintptr_t)1);
+}
+
+static int
+IS_REC_DELAY(trie_node *r)
+{ return (uintptr_t)r & 0x1;
+}
+
 static int
 answer_is_conditional(trie_node *answer)
 { delay_info *di;
@@ -492,9 +560,11 @@ answer_delay_info(worklist *wl, trie_node *answer, int create)
   { return NULL;
   } else if ( (di=malloc(sizeof(*di))) )
   { di->variant = wl->table->data.variant;
+    di->has_share_records = FALSE;
     initBuffer(&di->delay_sets);
     initBuffer(&di->delays);
     answer->data.delayinfo = di;
+    wl->undefined++;
 
     return di;
   } else
@@ -503,9 +573,19 @@ answer_delay_info(worklist *wl, trie_node *answer, int create)
 }
 
 
-static void
+void
 destroy_delay_info(delay_info *di)
-{ discardBuffer(&di->delay_sets);
+{ if ( di->has_share_records )
+  { delay *d = baseBuffer(&di->delays, delay);
+    delay *z = topBuffer(&di->delays, delay);
+
+    for(; d < z; d++)		/* keep a flag to see whether we have these */
+    { if ( IS_REC_DELAY(d->answer) )
+	PL_erase(UNREC_DELAY(d->answer));
+    }
+  }
+
+  discardBuffer(&di->delay_sets);
   discardBuffer(&di->delays);
   free(di);
 }
@@ -626,23 +706,6 @@ simplify_delay_set(delay_info *di, delay_set *ds)
   return FALSE;
 }
 
-static inline trie_node *
-REC_DELAY(record_t r)
-{ return (trie_node*)(((uintptr_t)r)|0x1);
-}
-
-static inline record_t
-UNREC_DELAY(trie_node *r)
-{ return (record_t)(((uintptr_t)r)&~(uintptr_t)1);
-}
-
-#ifndef NDEBUG					/* only used in assert() */
-static int
-IS_REC_DELAY(trie_node *r)
-{ return (uintptr_t)r & 0x1;
-}
-#endif
-
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 The delay list (`delays`) is a  list   of  delayed positive and negative
@@ -682,6 +745,7 @@ retry:
     if ( (di=answer->data.delayinfo) )
     { answer->data.delayinfo = NULL;
       destroy_delay_info(di);
+      wl->undefined--;
       DEBUG(MSG_TABLING_SIMPLIFY,
 	    Sdprintf("Unconditional answer after conditional\n"));
     }
@@ -807,6 +871,7 @@ retry:
 	if ( r )
 	{ if ( !add_to_delay_set(di, ds, NULL, REC_DELAY(r)) )
 	    goto nomem;
+	  di->has_share_records = TRUE;
 	} else
 	{ return UDL_FALSE;
 	}
@@ -1064,12 +1129,14 @@ remove_conditional_answer(spf_agenda *agenda, trie_node *answer)
 { delay_info *di = answer->data.delayinfo;
   trie *at = symbol_trie(di->variant->value);
   worklist *wl = at->data.worklist;
+
   assert(wl->magic == WORKLIST_MAGIC);
 
   DEBUG(MSG_TABLING_SIMPLIFY,
 	print_delay("    Removing conditional answer", di->variant, answer));
 
   answer->data.delayinfo = NULL;
+  destroy_delay_info(di);
   trie_delete(at, answer, FALSE);		/* cannot prune as may be */
   agenda->done++;				/* in worklist delay lists */
   wl->undefined--;
@@ -1127,7 +1194,9 @@ propagate_to_answer(spf_agenda *agenda, worklist *wl,
 	  { int res;
 
 	    DEBUG(MSG_TABLING_SIMPLIFY,
-		  Sdprintf("   found\n"));
+		  Sdprintf("   found (SCC=%zd, simplifications = %zd)\n",
+			   pointerToInt(wl->component),
+			   wl->component->simplifications));
 
 	    if ( d->answer == NULL )
 	    { if ( result == FALSE &&
@@ -1139,6 +1208,7 @@ propagate_to_answer(spf_agenda *agenda, worklist *wl,
 	    }
 
 	    found = TRUE;
+	    wl->component->simplifications++;
 
 	    if ( res )			/* remove member from conjunction */
 	    { d->variant = DV_DELETED;
@@ -1198,15 +1268,22 @@ simplify_component(tbl_component *scc)
   worklist **wlp0 = baseBuffer(&scc->created_worklists->members, worklist*);
   worklist **top  = topBuffer(&scc->created_worklists->members, worklist*);
   worklist **wlp;
-  int undefined;
-  int tcount = 0;
+  int undefined, pass;
+#ifndef O_AC_EAGER
+  size_t simplified0 = scc->simplifications;
+#endif
 
   DEBUG(MSG_TABLING_SIMPLIFY,
-	Sdprintf("Simplifying SCC %zd\n", pointerToInt(scc)));
+	{ GET_LD
+	  term_t t = PL_new_term_ref();
+	  unify_trie_term(scc->leader->data.variant, t PASS_LD);
+	  Sdprintf("Simplifying SCC %zd; leader = ", pointerToInt(scc));
+	  PL_write_term(Serror, t, 999, PL_WRT_NEWLINE);
+	});
 
   init_spf_agenda(&agenda);
 
-  for(;;)
+  for(pass=0; ;pass++)
   { int count = 0;
 
     undefined = 0;
@@ -1214,6 +1291,9 @@ simplify_component(tbl_component *scc)
     for(wlp = wlp0; wlp < top; wlp++)
     { worklist *wl = *wlp;
 
+      if ( pass == 0 )
+	clean_worklist(wl);
+
       if ( wl->negative &&
 	   wl->neg_delayed &&
 	   wl->table->value_count == 0 &&
@@ -1258,19 +1338,31 @@ simplify_component(tbl_component *scc)
 	undefined++;
     }
 
-    tcount += count;
-
     if ( count == 0 || undefined == 0 )
       break;
   }
 
   exit_spf_agenda(&agenda);
 
+#ifndef O_AC_EAGER
+  if ( (simplified0 != scc->simplifications) )
+  { size_t cnt = scc->simplifications - simplified0 ;
+    tbl_component *c;
+
+    for(c = scc->parent; c; c = c->parent)
+      c->simplifications += cnt;
+  }
+#endif
+
   /* DSW: there cannot be any "uncovering" of a positive loop if there
    * was no simplification
    */
 
-  if ( undefined && tcount )
+  DEBUG(MSG_TABLING_SIMPLIFY,
+	Sdprintf("Simplified SCC %zd; undefined = %d; simplifications: %zd\n",
+		 pointerToInt(scc), undefined, scc->simplifications));
+
+  if ( undefined && scc->simplifications )
     return answer_completion(scc);
   else
     return TRUE;
@@ -1294,9 +1386,8 @@ one positive dependency.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
-call_answer_completion(worklist *wl ARG_LD)
+call_answer_completion(trie *atrie ARG_LD)
 { fid_t fid;
-  trie *atrie = wl->table;
 
   if ( (fid = PL_open_foreign_frame()) )
   { static predicate_t pred = NULL;
@@ -1310,7 +1401,7 @@ call_answer_completion(worklist *wl ARG_LD)
 
     DEBUG(MSG_TABLING_AC,
 	  { term_t t = PL_new_term_ref();
-	    unify_trie_term(wl->table->data.variant, t PASS_LD);
+	    unify_trie_term(atrie->data.variant, t PASS_LD);
 	    Sdprintf("Calling answer completion for: ");
 	    PL_write_term(Serror, t, 999, PL_WRT_NEWLINE);
 	  });
@@ -1338,6 +1429,7 @@ one answer that  is  undefined  and   has  a  condition  containing only
 positive delay elements.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#ifdef O_AC_EAGER
 static int
 has_positive_dl(trie_node *n)
 { delay_info *di;
@@ -1379,13 +1471,12 @@ is_ac_candidate_wl(worklist *wl)
 
   return FALSE;
 }
+#endif /*O_AC_EAGER*/
 
 
 static int
 answer_completion(tbl_component *scc)
 { GET_LD
-  worklist **wlp = baseBuffer(&scc->created_worklists->members, worklist*);
-  worklist **top = topBuffer(&scc->created_worklists->members, worklist*);
 
 #ifdef O_DEBUG
   if ( DEBUGGING(TABLING_NO_AC) )
@@ -1393,18 +1484,25 @@ answer_completion(tbl_component *scc)
 #endif
 
   if ( LD->tabling.in_answer_completion )
-    return TRUE;				/* Warn? */
+    return TRUE;			/* not recursive! */
+
+#ifdef O_AC_EAGER
+  worklist **wlp = baseBuffer(&scc->created_worklists->members, worklist*);
+  worklist **top = topBuffer(&scc->created_worklists->members, worklist*);
 
   for(; wlp < top; wlp++)
   { worklist *wl = *wlp;
 
     if ( is_ac_candidate_wl(wl) )
-    { if ( !call_answer_completion(wl PASS_LD) )
+    { if ( !call_answer_completion(wl->table PASS_LD) )
 	return FALSE;
     }
   }
 
   return TRUE;
+#else
+  return call_answer_completion(scc->leader PASS_LD);
+#endif /*O_AC_EAGER*/
 }
 
 /** '$tbl_force_truth_value'(+AnswerNode, +Value, -Count)
@@ -1489,6 +1587,9 @@ PRED_IMPL("$tbl_set_answer_completed", 1, tbl_set_answer_completed, 0)
       return TRUE;
     }
 
+    if ( wl == WL_COMPLETE )
+      return TRUE;
+
     return PL_permission_error("set_answer_complete", "trie", A1);
   }
 
@@ -1580,10 +1681,20 @@ release_variant_table_node(trie *variant_table, trie_node *node)
 
   if ( node->value )
   { trie *vtrie = symbol_trie(node->value);
+    record_t r;
+    worklist *wl;
+
+    if ( WL_IS_WORKLIST(wl=vtrie->data.worklist) )
+      free_worklist(wl);
 
     assert(vtrie->data.variant == node);
     vtrie->data.variant = NULL;
     vtrie->data.worklist = NULL;
+    if ( (r=vtrie->data.skeleton) )
+    { vtrie->data.skeleton = 0;
+      PL_erase(r);
+    }
+
     trie_empty(vtrie);
   }
 }
@@ -1614,6 +1725,7 @@ get_variant_table(term_t t, int create ARG_LD)
       node->value = trie_symbol(vt);
       vt->data.variant = node;
       vt->alloc_pool = &LD->tabling.node_pool;
+      ATOMIC_INC(&variants->value_count);
       return vt;
     } else
       return NULL;
@@ -1627,7 +1739,7 @@ get_variant_table(term_t t, int create ARG_LD)
 void
 clearThreadTablingData(PL_local_data_t *ld)
 { reset_global_worklist(ld->tabling.component);
-  reset_newly_created_worklists(ld->tabling.component);
+  reset_newly_created_worklists(ld->tabling.component, WLFS_KEEP_COMPLETE);
   clear_variant_table(ld);
 }
 
@@ -1827,19 +1939,47 @@ free_worklist(worklist *wl)
 }
 
 
-static int
-worklist_negative(worklist *wl)
-{ wl->negative = TRUE;
-  if ( wl->component->neg_status == SCC_NEG_NONE )
-    wl->component->neg_status = SCC_NEG_DELAY;
+static void
+clean_worklist(worklist *wl)
+{ cluster *c, *next;
 
-  return TRUE;
+  wl->riac = NULL;
+  if ( wl->head )
+  { for(c=wl->head; c; c = next)
+    { next = c->next;
+      free_cluster(c);
+    }
+    wl->head = wl->tail = NULL;
+  }
+
+  if ( wl->free_clusters )
+  { for(c=wl->free_clusters; c; c = next)
+    { next = c->next;
+      free_cluster(c);
+    }
+    wl->free_clusters = NULL;
+  }
 }
 
+
+static void
+complete_worklist(worklist *wl)
+{ clean_worklist(wl);
+
+  wl->completed = TRUE;
+}
+
+
 static int
-worklist_is_false(worklist *wl)
-{ assert(wl->negative);
-  return (wl->neg_delayed) && !wl->has_answers;
+worklist_negative(worklist *wl)
+{ if ( !wl->negative )
+  { wl->negative = TRUE;
+    add_delay_worklist(wl);
+    if ( wl->component->neg_status == SCC_NEG_NONE )
+      wl->component->neg_status = SCC_NEG_DELAY;
+  }
+
+  return TRUE;
 }
 
 
@@ -1930,9 +2070,7 @@ static int
 wkl_add_answer(worklist *wl, trie_node *node ARG_LD)
 { potentially_add_to_global_worklist(wl PASS_LD);
 
-  if ( answer_is_conditional(node) )
-    wl->undefined++;
-  else
+  if ( !answer_is_conditional(node) )
     wl->has_answers = TRUE;
 
   if ( wl->head && wl->head->type == CLUSTER_ANSWERS )
@@ -1999,13 +2137,17 @@ unify_table_status(term_t t, trie *trie, int merge ARG_LD)
 { worklist *wl = trie->data.worklist;
 
   if ( WL_IS_WORKLIST(wl) )
-  { if ( merge && wl->component != LD->tabling.component )
+  { if ( wl->completed )
+      return PL_unify_atom(t, ATOM_complete);
+
+    if ( merge && wl->component != LD->tabling.component )
     { DEBUG(MSG_TABLING_WORK,
 	    Sdprintf("Merging into %p (current = %p)\n",
 		     wl->component, LD->tabling.component));
       merge_component(wl->component);
       LD->tabling.component = wl->component;
     }
+
     return PL_unify_pointer(t, wl);
   }
   if ( wl == WL_COMPLETE )
@@ -2075,9 +2217,8 @@ get_scc(term_t t, tbl_component **cp)
 }
 
 static int
-get_worklist(term_t t, worklist **wlp)
-{ GET_LD
-  void *ptr;
+get_worklist(term_t t, worklist **wlp ARG_LD)
+{ void *ptr;
 
   if ( PL_get_pointer(t, &ptr) )
   { worklist *wl = ptr;
@@ -2178,7 +2319,13 @@ PRED_IMPL("$tbl_destroy_table", 1, tbl_destroy_table, 0)
     { trie *vtrie = get_trie_form_node(table->data.variant);
 
       if ( vtrie == LD->tabling.variant_table )
-      { prune_node(vtrie, table->data.variant);
+      { record_t r;
+
+	if ( (r=table->data.skeleton) )
+	{ table->data.skeleton = 0;
+	  PL_erase(r);
+	}
+	trie_delete(vtrie, table->data.variant, TRUE);
 	return TRUE;
       }
 
@@ -2206,9 +2353,14 @@ PRED_IMPL("$tbl_pop_worklist", 2, tbl_pop_worklist, 0)
     if ( (wl=pop_worklist(scc PASS_LD)) )
       return PL_unify_pointer(A2, wl);
 
-    if ( scc->neg_status != SCC_NEG_NONE &&
-	 (wl=negative_worklist(scc PASS_LD)) )
-      return PL_unify_pointer(A2, wl);
+    if (
+#ifndef O_AC_EAGER
+	  scc->simplifications ||
+#endif
+	  scc->neg_status != SCC_NEG_NONE )
+    { if ( (wl=negative_worklist(scc PASS_LD)) )
+	return PL_unify_pointer(A2, wl);
+    }
   }
 
   return FALSE;
@@ -2226,7 +2378,7 @@ PRED_IMPL("$tbl_wkl_add_answer", 4, tbl_wkl_add_answer, 0)
 { PRED_LD
   worklist *wl;
 
-  if ( get_worklist(A1, &wl) )
+  if ( get_worklist(A1, &wl PASS_LD) )
   { Word kp;
     trie_node *node;
     int rc;
@@ -2287,7 +2439,7 @@ PRED_IMPL("$tbl_wkl_mode_add_answer", 4, tbl_wkl_mode_add_answer, 0)
 { PRED_LD
   worklist *wl;
 
-  if ( get_worklist(A1, &wl) )
+  if ( get_worklist(A1, &wl PASS_LD) )
   { Word kp;
     trie_node *node;
     int rc;
@@ -2366,9 +2518,10 @@ PRED_IMPL("$tbl_wkl_add_suspension", 2, tbl_wkl_add_suspension, 0)
 
 static
 PRED_IMPL("$tbl_wkl_make_follower", 1, tbl_wkl_make_follower, 0)
-{ worklist *wl;
+{ PRED_LD
+  worklist *wl;
 
-  if ( get_worklist(A1, &wl) )
+  if ( get_worklist(A1, &wl PASS_LD) )
   { cluster *scp = NULL;
     cluster *acp = NULL;
     cluster *c, *next;
@@ -2404,7 +2557,8 @@ PRED_IMPL("$tbl_wkl_make_follower", 1, tbl_wkl_make_follower, 0)
     if ( scp )
       wkl_append_right(wl, scp);
 
-    DEBUG(0, assert(is_in_global_worklist(wl)));
+    if ( acp && scp )
+      add_global_worklist(wl);
 
     return TRUE;
   }
@@ -2420,9 +2574,10 @@ PRED_IMPL("$tbl_wkl_make_follower", 1, tbl_wkl_make_follower, 0)
 
 static
 PRED_IMPL("$tbl_wkl_done", 1, tbl_wkl_done, 0)
-{ worklist *wl;
+{ PRED_LD
+  worklist *wl;
 
-  return get_worklist(A1, &wl) && worklist_work_done(wl);
+  return get_worklist(A1, &wl PASS_LD) && worklist_work_done(wl);
 }
 
 /** '$tbl_wkl_negative'(+Worklist) is semidet.
@@ -2432,23 +2587,34 @@ PRED_IMPL("$tbl_wkl_done", 1, tbl_wkl_done, 0)
 
 static
 PRED_IMPL("$tbl_wkl_negative", 1, tbl_wkl_negative, 0)
-{ worklist *wl;
+{ PRED_LD
+  worklist *wl;
 
-  return get_worklist(A1, &wl) && worklist_negative(wl);
+  return get_worklist(A1, &wl PASS_LD) && worklist_negative(wl);
 }
 
 
 /** '$tbl_tbl_wkl_is_false'(+Worklist) is semidet.
  *
- * True if the worklist is is a negative node that is true (has no
- * solutions)
+ * True if the worklist is is  a  negative   node  that  is true (has no
+ * definite solutions). This is used at   the end of negation_suspend/4,
+ * after we delayed the negation. This  means   we  must fail if we have
+ * definite answers, but succeed if  there   are  conditional answers as
+ * simplification has not yet been done.
  */
 
 static
 PRED_IMPL("$tbl_wkl_is_false", 1, tbl_wkl_is_false, 0)
-{ worklist *wl;
+{ PRED_LD
+  worklist *wl;
 
-  return get_worklist(A1, &wl) && worklist_is_false(wl);
+  if ( get_worklist(A1, &wl PASS_LD) )
+  { assert(wl->negative);
+
+    return wl->neg_delayed && !wl->has_answers;
+  }
+
+  return FALSE;
 }
 
 /** '$tbl_wkl_answer_trie'(+Worklist, -Trie) is det.
@@ -2461,7 +2627,7 @@ PRED_IMPL("$tbl_wkl_answer_trie", 2, tbl_wkl_answer_trie, 0)
 { GET_LD
   worklist *wl;
 
-  return ( get_worklist(A1, &wl) &&
+  return ( get_worklist(A1, &wl PASS_LD) &&
 	   PL_unify_atom(A2, wl->table->symbol) );
 }
 
@@ -2514,8 +2680,7 @@ unify_arg_term(term_t a, Word v ARG_LD)
 static int
 unify_dependency(term_t a0, term_t dependency,
 		 worklist *wl, trie_node *answer ARG_LD)
-{ if ( ensureTrailSpace(4) &&
-       ensureGlobalSpace(6, ALLOW_GC) )
+{ if ( likely(ensureStackSpace__LD(6, 5, ALLOW_GC PASS_LD)) )
   { Word dp = valTermRef(dependency);
     Functor f;
 
@@ -2631,7 +2796,7 @@ PRED_IMPL("$tbl_wkl_work", 8, tbl_wkl_work, PL_FA_NONDETERMINISTIC)
   { case FRG_FIRST_CALL:
     { worklist *wl;
 
-      if ( get_worklist(A1, &wl) )
+      if ( get_worklist(A1, &wl PASS_LD) )
       { cluster *acp, *scp;
 
 	if ( (acp=wl->riac) && (scp=acp->next) )
@@ -2697,13 +2862,14 @@ next:
       term_t suspension = av+0;
       term_t modeargs   = av+1;
 
-      if ( an == NULL && !IS_TNOT(sr) )
-      { if ( advance_wkl_state(state) )
-	  goto next;
-	goto out_fail;
-      }
-      if ( an != NULL && IS_TNOT(sr) &&
-	   answer_is_conditional(an) )
+      /* Ignore (1) dummy restart for delayed negation,
+       *        (2) conditional answer for tnot
+       *        (3) removed answer due to simplification
+       */
+
+      if ( (an == NULL && !IS_TNOT(sr)) ||
+	   (an != NULL && IS_TNOT(sr) && answer_is_conditional(an)) ||
+	   (an != NULL && an->value == 0) )
       { if ( advance_wkl_state(state) )
 	  goto next;
 	goto out_fail;
@@ -2776,6 +2942,21 @@ PRED_IMPL("$tbl_variant_table", 4, tbl_variant_table, 0)
 }
 
 
+static
+PRED_IMPL("$tbl_existing_variant_table", 4, tbl_existing_variant_table, 0)
+{ PRED_LD
+  trie *trie;
+
+  if ( (trie=get_variant_table(A1, FALSE PASS_LD)) )
+  { return ( _PL_unify_atomic(A2, trie->symbol) &&
+	     unify_table_status(A3, trie, TRUE PASS_LD)  &&
+	     unify_skeleton(trie, A1, A4 PASS_LD) );
+  }
+
+  return FALSE;
+}
+
+
 static
 PRED_IMPL("$tbl_variant_table", 1, tbl_variant_table, 0)
 { PRED_LD
@@ -2834,9 +3015,14 @@ PRED_IMPL("$tbl_table_complete_all", 1, tbl_table_complete_all, 0)
 	      PL_write_term(Serror, t, 999, PL_WRT_NEWLINE);
 	    });
 
-      trie->data.worklist = WL_COMPLETE;
+      if ( !wl->undefined && isEmptyBuffer(&wl->delays) )
+      { free_worklist(wl);
+	trie->data.worklist = WL_COMPLETE;
+      } else
+      { complete_worklist(wl);
+      }
     }
-    reset_newly_created_worklists(c);
+    reset_newly_created_worklists(c, WLFS_FREE_NONE);
     c->status = SCC_COMPLETED;
 
     if ( c->parent && LD->tabling.component == c )
@@ -2896,7 +3082,7 @@ PRED_IMPL("$tbl_table_discard_all", 1, tbl_table_discard_all, 0)
 
 	prune_node(LD->tabling.variant_table, trie->data.variant);
       }
-      reset_newly_created_worklists(c);
+      reset_newly_created_worklists(c, WLFS_KEEP_COMPLETE);
     }
     reset_global_worklist(c);
     // FIXME: just pop?
@@ -2922,6 +3108,7 @@ PRED_IMPL("$tbl_create_subcomponent", 2, tbl_create_subcomponent, 0)
       LD->tabling.component = new_component();
     else
       LD->tabling.component->status = SCC_ACTIVE;
+    LD->tabling.component->leader = leader;
     return PL_unify_pointer(A1, LD->tabling.component);
   }
 
@@ -3088,6 +3275,19 @@ PRED_IMPL("$tbl_scc_data", 2, tbl_scc_data, 0)
 }
 
 
+static int
+uc_put_trie_value(term_t t, trie_node *an ARG_LD)
+{ static atom_t anull;
+
+  if ( !anull )
+    anull = PL_new_atom("NULL");
+
+  if ( !an || an->value )
+    return tbl_put_trie_value(t, an PASS_LD);
+  else
+    return PL_put_atom(t, anull);
+}
+
 static int
 unify_cluster(term_t t, cluster *c, int is_riac)
 { GET_LD
@@ -3110,11 +3310,13 @@ unify_cluster(term_t t, cluster *c, int is_riac)
     term_t modeav = PL_new_term_ref();
 
     for(; ap < top; ap++)
-    { if ( !PL_put_variable(answer) ||
+    { trie_node *an = *ap;
+
+      if ( !PL_put_variable(answer) ||
 	   !PL_unify_list(tail, head, tail) ||
-	   !tbl_unify_trie_term(*ap, answer PASS_LD) ||
-	   !tbl_put_trie_value(modeav, *ap PASS_LD) ||
-	   !PL_unify_term(head, PL_FUNCTOR_CHARS, "-", 2,
+	   !tbl_unify_trie_term(an, answer PASS_LD) ||
+	   !uc_put_trie_value(modeav, an PASS_LD) ||
+	   !PL_unify_term(head, PL_FUNCTOR, FUNCTOR_minus2,
 			          PL_TERM, answer, PL_TERM, modeav) )
 	return FALSE;
     }
@@ -3159,7 +3361,7 @@ PRED_IMPL("$tbl_worklist_data", 2, tbl_worklist_data, 0)
 { PRED_LD
   worklist *wl;
 
-  if ( get_worklist(A1, &wl) )
+  if ( get_worklist(A1, &wl PASS_LD) )
   { term_t av = PL_new_term_refs(5);
     term_t t = PL_new_term_ref();
     static functor_t f = 0;
@@ -3390,6 +3592,46 @@ PRED_IMPL("$tbl_answer_update_dl", 2, tbl_answer_update_dl,
 }
 
 
+/** '$tbl_implementation'(:G0, -G) is det.
+ *
+ * Find location where G is actually defined and raise an error of the
+ * predicate is not tabled.
+ */
+
+static
+PRED_IMPL("$tbl_implementation", 2, tbl_implementation, PL_FA_TRANSPARENT)
+{ PRED_LD
+  Module m = NULL;
+  term_t t = PL_new_term_ref();
+  functor_t f = 0;
+  Procedure proc;
+  Definition def;
+
+  if ( !PL_strip_module(A1, &m, t) )
+    return FALSE;
+  if ( !PL_get_functor(t, &f) )
+    return PL_type_error("callable", A1);
+  if ( !(proc = resolveProcedure(f, m)) )
+    return FALSE;				/* should not happen */
+
+  if ( !isDefinedProcedure(proc) )
+    trapUndefined(getProcDefinition(proc) PASS_LD);
+  def = getProcDefinition(proc);
+
+  if ( false(def, P_TABLED) )
+  { return PL_error(NULL, 0, NULL, ERR_PERMISSION_PROC,
+		    ATOM_tnot, ATOM_non_tabled_procedure, proc);
+  }
+
+  if ( def->module == m )
+  { return PL_unify(A2, A1);
+  } else
+  { return PL_unify_term(A2, PL_FUNCTOR, FUNCTOR_colon2,
+			       PL_ATOM, def->module->name,
+			       PL_TERM, t);
+  }
+}
+
 
 
 		 /*******************************
@@ -3397,6 +3639,7 @@ PRED_IMPL("$tbl_answer_update_dl", 2, tbl_answer_update_dl,
 		 *******************************/
 
 #define NDET PL_FA_NONDETERMINISTIC
+#define META PL_FA_TRANSPARENT
 
 BeginPredDefs(tabling)
   PRED_DEF("$tbl_new_worklist",		2, tbl_new_worklist,	     0)
@@ -3411,6 +3654,7 @@ BeginPredDefs(tabling)
   PRED_DEF("$tbl_wkl_answer_trie",	2, tbl_wkl_answer_trie,      0)
   PRED_DEF("$tbl_wkl_work",		8, tbl_wkl_work,          NDET)
   PRED_DEF("$tbl_variant_table",	4, tbl_variant_table,	     0)
+  PRED_DEF("$tbl_existing_variant_table", 4, tbl_existing_variant_table, 0)
   PRED_DEF("$tbl_variant_table",        1, tbl_variant_table,        0)
   PRED_DEF("$tbl_table_status",		4, tbl_table_status,	     0)
   PRED_DEF("$tbl_table_complete_all",	1, tbl_table_complete_all,   0)
@@ -3434,4 +3678,5 @@ BeginPredDefs(tabling)
   PRED_DEF("$tbl_force_truth_value",    3, tbl_force_truth_value,    0)
   PRED_DEF("$tbl_set_answer_completed", 1, tbl_set_answer_completed, 0)
   PRED_DEF("$tbl_is_answer_completed",  1, tbl_is_answer_completed,  0)
+  PRED_DEF("$tbl_implementation",       2, tbl_implementation,    META)
 EndPredDefs
