@@ -35,6 +35,7 @@
 
 #include "pl-incl.h"
 #include "pl-trie.h"
+#include "pl-tabling.h"
 #include "pl-indirect.h"
 #define AC_TERM_WALK_POP 1
 #include "pl-termwalk.c"
@@ -60,12 +61,23 @@ TODO
   - Make trie_gen/3 take the known prefix into account
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define RESERVED_TRIE_VAL(n) (((word)(-(intptr_t)n)<<LMASK_BITS) | TAG_VAR)
+#define RESERVED_TRIE_VAL(n) (((word)((uintptr_t)n)<<LMASK_BITS) | \
+			      TAG_VAR|STG_LOCAL)
 #define TRIE_ERROR_VAL       RESERVED_TRIE_VAL(1)
-#define TRIE_KEY_POP         RESERVED_TRIE_VAL(2)
+#define TRIE_KEY_POP(n)      RESERVED_TRIE_VAL(10+(n))
+
+#define IS_TRIE_KEY_POP(w)   ((tagex(w) == (TAG_VAR|STG_LOCAL) && \
+			       ((w)>>LMASK_BITS) > 10) ? ((w)>>LMASK_BITS) - 10 \
+						       : 0)
 
 #define NVARS_FAST 100
 
+#ifdef O_TRIE_STATS
+#define TRIE_STAT_INC(t, v) ATOMIC_INC(&t->stats.v)
+#else
+#define TRIE_STAT_INC(t, v) ((void)0)
+#endif
+
 /* Will eventually be shared in pl-wam.c */
 typedef enum
 { uread = 0,				/* Unification in read-mode */
@@ -183,6 +195,7 @@ trie_create(void)
   if ( (trie = PL_malloc(sizeof(*trie))) )
   { memset(trie, 0, sizeof(*trie));
     trie->magic = TRIE_MAGIC;
+    trie->node_count = 1;		/* the root */
 
     return trie;
   } else
@@ -281,17 +294,20 @@ next:
   release_key(n->key);
   if ( n->value )
     release_value(n->value);
+  if ( n->data.delayinfo )		/* hacky */
+    destroy_delay_info(n->data.delayinfo);
 
-  if ( children.any &&
-       COMPARE_AND_SWAP(&n->children.any, children.any, NULL) )
-  { if ( dealloc )
-    { ATOMIC_DEC(&trie->node_count);
-      if ( trie->alloc_pool )
-	ATOMIC_SUB(&trie->alloc_pool->size, sizeof(trie_node));
-      PL_free(n);
-    }
+  if ( dealloc )
+  { ATOMIC_DEC(&trie->node_count);
+    if ( trie->alloc_pool )
+      ATOMIC_SUB(&trie->alloc_pool->size, sizeof(trie_node));
+    PL_free(n);
+  } else
+  { n->children.any = NULL;
+  }
 
-    switch( children.any->type )
+  if ( children.any )
+  { switch( children.any->type )
     { case TN_KEY:
       { n = children.key->child;
         PL_free(children.key);
@@ -377,11 +393,16 @@ insert_child(trie *trie, trie_node *n, word key ARG_LD)
 	  } else
 	  { trie_children_hashed *hnode = PL_malloc(sizeof(*hnode));
 
-	    hnode->type  = TN_HASHED;
-	    hnode->table = newHTable(4);
+	    hnode->type     = TN_HASHED;
+	    hnode->table    = newHTable(4);
+	    hnode->var_keys = 0;
 	    addHTable(hnode->table, (void*)children.key->key,
 				    children.key->child);
 	    addHTable(hnode->table, (void*)key, (void*)new);
+	    if ( tagex(children.key->key) == TAG_VAR )
+	      hnode->var_keys++;
+	    if ( tagex(new->key) == TAG_VAR )
+	      hnode->var_keys++;
 
 	    if ( COMPARE_AND_SWAP(&n->children.hash, children.hash, hnode) )
 	    { PL_free(children.any);		/* TBD: Safely free */
@@ -400,6 +421,8 @@ insert_child(trie *trie, trie_node *n, word key ARG_LD)
 
 	  if ( new == old )
 	  { new->parent = n;
+	    if ( tagex(new->key) == TAG_VAR )
+	      ATOMIC_INC(&children.hash->var_keys);
 	  } else
 	  { destroy_node(trie, new);
 	  }
@@ -476,19 +499,26 @@ trie_lookup(trie *trie, trie_node **nodep, Word k, int add ARG_LD)
   trie_node *node = &trie->root;
   size_t var_number = 0;
   int rc = TRUE;
-  int compounds = 0;
+  size_t compounds = 0;
+
+  TRIE_STAT_INC(trie, lookups);
 
   initTermAgenda_P(&agenda, 1, k);
   while( node )
   { Word p;
     word w;
+    size_t popn;
 
     if ( !(p=nextTermAgenda_P(&agenda)) )
       break;
-    if ( p == AC_TERM_POP )
-    { if ( !(node = follow_node(trie, node, TRIE_KEY_POP, add PASS_LD)) )
-	break;
-      continue;
+    if ( (popn = IS_AC_TERM_POP(p)) )
+    { compounds -= popn;
+      if ( compounds > 0 )
+      { if ( !(node = follow_node(trie, node, TRIE_KEY_POP(popn), add PASS_LD)) )
+	  break;
+	continue;
+      } else
+	break;				/* finished toplevel */
     }
 
     w = *p;
@@ -508,7 +538,7 @@ trie_lookup(trie *trie, trie_node **nodep, Word k, int add ARG_LD)
       { Functor f = valueTerm(w);
         size_t arity = arityFunctor(f->definition);
 
-	if ( add && ++compounds == 1000 && !is_acyclic(p PASS_LD) )
+	if ( ++compounds == 1000 && add && !is_acyclic(p PASS_LD) )
 	{ rc = TRIE_LOOKUP_CYCLIC;
 	  prune_error(trie, node PASS_LD);
 	  node = NULL;
@@ -596,7 +626,7 @@ get_trie_form_node(trie_node *node)
 int
 is_ground_trie_node(trie_node *node)
 { for( ; node->parent; node = node->parent )
-  { if ( tag(node->key) == TAG_VAR && node->key != TRIE_KEY_POP )
+  { if ( tagex(node->key) == TAG_VAR )
       return FALSE;
   }
 
@@ -1243,150 +1273,182 @@ find_var(ukey_state *state, size_t index)
 
 static int
 unify_key(ukey_state *state, word key ARG_LD)
-{ Word p;
+{ Word p = state->ptr;
 
-  if ( key == TRIE_KEY_POP )
-  { Word wp = *--aTop;
-    state->umode = ((int)(uintptr_t)wp & uwrite);
-    state->ptr   = (Word)((intptr_t)wp&~uwrite);
+  switch(tagex(key))
+  { case TAG_VAR|STG_LOCAL:			/* RESERVED_TRIE_VAL */
+    { size_t popn = IS_TRIE_KEY_POP(key);
+      Word wp;
 
-    DEBUG(MSG_TRIE_PUT_TERM,
-	  Sdprintf("U Popped %zd, mode=%d\n", state->ptr-gBase, state->umode));
-    return TRUE;
-  }
+      assert(popn);
+      aTop -= popn;
+      wp = *aTop;
+      state->umode = ((int)(uintptr_t)wp & uwrite);
+      state->ptr   = (Word)((intptr_t)wp&~uwrite);
 
-  p = state->ptr;
-  if ( state->umode == uread )
-    deRef(p);
+      DEBUG(MSG_TRIE_PUT_TERM,
+	    Sdprintf("U Popped(%zd) %zd, mode=%d\n",
+		     popn, state->ptr-gBase, state->umode));
+      return TRUE;
+    }
+    case TAG_ATOM|STG_GLOBAL:			/* functor */
+    { size_t arity = arityFunctor(key);
 
-  if ( tagex(key) == (TAG_ATOM|STG_GLOBAL) )
-  { size_t arity = arityFunctor(key);
+      DEBUG(MSG_TRIE_PUT_TERM,
+	    Sdprintf("U Pushed %s %zd, mode=%d\n",
+		     functorName(key), state->ptr+1-gBase, state->umode));
+      pushArgumentStack((Word)((intptr_t)(p + 1)|state->umode));
 
-    DEBUG(MSG_TRIE_PUT_TERM,
-	  Sdprintf("U Pushed %s %zd, mode=%d\n",
-		   functorName(key), state->ptr+1-gBase, state->umode));
-    pushArgumentStack((Word)((intptr_t)(state->ptr + 1)|state->umode));
+      if ( state->umode == uwrite )
+      { Word t;
 
-    if ( state->umode == uwrite )
-    { Word t;
-
-      if ( (t=allocGlobalNoShift(arity+1)) )
-      { t[0] = key;
-	*p = consPtr(t, TAG_COMPOUND|STG_GLOBAL);
-	state->ptr = &t[1];
-	return TRUE;
-      } else
-	return GLOBAL_OVERFLOW;
-    } else
-    { if ( canBind(*p) )
-      { state->umode = uwrite;
-
-	if ( isAttVar(*p) )
-	{ Word t;
-	  word w;
-	  size_t i;
-
-	  if ( (t=allocGlobalNoShift(arity+1)) )
-	  { if ( !hasGlobalSpace(0) )
-	      return overflowCode(0);
-	    w = consPtr(&t[0], TAG_COMPOUND|STG_GLOBAL);
-	    t[0] = key;
-	    for(i=0; i<arity; i++)
-	      setVar(t[i+1]);
-	    assignAttVar(p, &w PASS_LD);
-	    state->ptr = &t[1];
-	    return TRUE;
-	  } else
-	    return GLOBAL_OVERFLOW;
-	} else
-	{ Word t;
-
-	  if ( (t=allocGlobalNoShift(arity+1)) )
-	  { if ( unlikely(tTop+1 >= tMax) )
-	      return  TRAIL_OVERFLOW;
-	    t[0] = key;
-	    Trail(p, consPtr(t, TAG_COMPOUND|STG_GLOBAL));
-	    state->ptr = &t[1];
-	    return TRUE;
-	  } else
-	    return GLOBAL_OVERFLOW;
-	}
-      } else if ( isTerm(*p) )
-      { Functor f = valueTerm(*p);
-
-	if ( f->definition == key )
-	{ state->ptr = &f->arguments[0];
+	if ( (t=allocGlobalNoShift(arity+1)) )
+	{ t[0] = key;
+	  *p = consPtr(t, TAG_COMPOUND|STG_GLOBAL);
+	  state->ptr = &t[1];
 	  return TRUE;
 	} else
-	  return FALSE;
+	  return GLOBAL_OVERFLOW;
       } else
-      { return FALSE;
-      }
+      { deRef(p);
+
+	if ( canBind(*p) )
+	{ state->umode = uwrite;
+
+	  if ( isAttVar(*p) )
+	  { Word t;
+	    word w;
+	    size_t i;
+
+	    if ( (t=allocGlobalNoShift(arity+1)) )
+	    { if ( !hasGlobalSpace(0) )
+		return overflowCode(0);
+	      w = consPtr(&t[0], TAG_COMPOUND|STG_GLOBAL);
+	      t[0] = key;
+	      for(i=0; i<arity; i++)
+		setVar(t[i+1]);
+	      assignAttVar(p, &w PASS_LD);
+	      state->ptr = &t[1];
+	      return TRUE;
+	    } else
+	      return GLOBAL_OVERFLOW;
+	  } else
+	  { Word t;
+
+	    if ( (t=allocGlobalNoShift(arity+1)) )
+	    { if ( unlikely(tTop+1 >= tMax) )
+		return  TRAIL_OVERFLOW;
+	      t[0] = key;
+	      Trail(p, consPtr(t, TAG_COMPOUND|STG_GLOBAL));
+	      state->ptr = &t[1];
+	      return TRUE;
+	    } else
+	      return GLOBAL_OVERFLOW;
+	  }
+	} else if ( isTerm(*p) )
+	{ Functor f = valueTerm(*p);
+
+	  if ( f->definition == key )
+	  { state->ptr = &f->arguments[0];
+	    return TRUE;
+	  } else
+	    return FALSE;
+	} else
+	{ return FALSE;
+	}
+      } /*uread*/
     }
-  } else if ( tag(key) == TAG_VAR )
-  { size_t index = (size_t)(key>>LMASK_BITS);
-    Word *v = find_var(state, index);
+    assert(0);
+    case TAG_VAR:
+    { size_t index = (size_t)(key>>LMASK_BITS);
+      Word *v = find_var(state, index);
 
-    DEBUG(MSG_TRIE_PUT_TERM,
-	  { char b1[64]; char b2[64];
-	    Sdprintf("var %zd at %s (v=%p, *v=%s)\n",
-		     index,
-		     print_addr(state->ptr,b1),
-		     v, print_addr(*v,b2));
-	  });
+      DEBUG(MSG_TRIE_PUT_TERM,
+	    { char b1[64]; char b2[64];
+	      Sdprintf("var %zd at %s (v=%p, *v=%s)\n",
+		       index,
+		       print_addr(state->ptr,b1),
+		       v, print_addr(*v,b2));
+	    });
 
-    if ( state->umode == uwrite )
-    { if ( !*v )
-      { setVar(*state->ptr);
-	*v = state->ptr;
+      if ( state->umode == uwrite )
+      { if ( !*v )
+	{ setVar(*state->ptr);
+	  *v = state->ptr;
+	} else
+	{ *state->ptr = makeRefG(*v);
+	}
       } else
-      { *state->ptr = makeRefG(*v);
-      }
-    } else
-    { if ( !*v )
-      { *v = state->ptr;
-      } else
-      { int rc;
+      { deRef(p);
 
-	if ( (rc=unify_ptrs(state->ptr, *v, ALLOW_RETCODE PASS_LD)) != TRUE )
-	  return rc;
+	if ( !*v )
+	{ *v = state->ptr;
+	} else
+	{ int rc;
+
+	  if ( (rc=unify_ptrs(state->ptr, *v, ALLOW_RETCODE PASS_LD)) != TRUE )
+	    return rc;
+	}
       }
+
+      break;
     }
+    assert(0);
+    case STG_GLOBAL|TAG_INTEGER:		/* indirect data */
+    case STG_GLOBAL|TAG_STRING:
+    case STG_GLOBAL|TAG_FLOAT:
+    { word w;
 
-    state->ptr++;
-    return TRUE;
-  } else
-  { word w;
-
-    DEBUG(MSG_TRIE_PUT_TERM,
-	  Sdprintf("%s at %s\n",
-		   print_val(key, NULL), print_addr(state->ptr,NULL)));
-
-    if ( isIndirect(key) )
-    { w = extern_indirect_no_shift(state->trie->indirects, key PASS_LD);
+      w = extern_indirect_no_shift(state->trie->indirects, key PASS_LD);
       if ( !w )
 	return GLOBAL_OVERFLOW;
-    } else
-      w = key;
 
-    if ( state->umode == uwrite )
-    { if ( isAtom(w) )
-	pushVolatileAtom(w);
-      *p = w;
-    } else if ( canBind(*p) )
-    { if ( isAtom(w) )
-	pushVolatileAtom(w);
+      if ( state->umode == uwrite )
+      { *p = w;
+      } else
+      { deRef(p);
 
-      if ( hasGlobalSpace(0) )
-	bindConst(p, w);
-      else
-	return overflowCode(0);
-    } else if ( *p != w )
-      return FALSE;
+	if ( canBind(*p) )
+	{ if ( hasGlobalSpace(0) )
+	    bindConst(p, w);
+	  else
+	    return overflowCode(0);
+	} else
+	{ if ( !equalIndirect(w, *p) )
+	    return FALSE;
+	}
+      }
 
-    state->ptr++;
-    return TRUE;
+      break;
+    }
+    case TAG_ATOM:
+      pushVolatileAtom(key);
+      /*FALLTHROUGH*/
+    case TAG_INTEGER:
+    { if ( state->umode == uwrite )
+      { *p = key;
+      } else
+      { deRef(p);
+
+	if ( canBind(*p) )
+	{ if ( hasGlobalSpace(0) )
+	    bindConst(p, key);
+	  else
+	    return overflowCode(0);
+	} else
+	{ if ( *p != key )
+	    return FALSE;
+	}
+      }
+      break;
+    }
+    default:
+      assert(0);
   }
+
+  state->ptr++;
+
+  return TRUE;
 }
 
 
@@ -1401,10 +1463,17 @@ typedef struct trie_choice
 
 typedef struct
 { trie        *trie;		/* trie we operate on */
-  int	       allocated;
+  int	       allocated;	/* If TRUE, the state is persistent */
   tmp_buffer   choicepoints;	/* Stack of trie state choicepoints */
 } trie_gen_state;
 
+typedef struct
+{ Word	       term;		/* Term we are descending */
+  int	       compound;	/* Initialized for compound */
+  int	       prune;		/* Use for pruning */
+  segstack     stack;		/* Stack for argument handling */
+  Word	       buffer[64];	/* Quick buffer for stack */
+} descent_state;
 
 static void
 init_trie_state(trie_gen_state *state, trie *trie)
@@ -1437,35 +1506,109 @@ clear_trie_state(trie_gen_state *state)
 }
 
 
+static void
+clear_descent_state(descent_state *dstate)
+{ if ( dstate->compound )
+    clearSegStack(&dstate->stack);
+}
+
+static int
+get_key(trie_gen_state *state, descent_state *dstate, word *key ARG_LD)
+{ Word p;
+
+  deRef2(dstate->term, p);
+
+  if ( canBind(*p) )
+  { return FALSE;
+  } else if ( isTerm(*p) )
+  { Functor f = valueTerm(*p);
+
+    *key = f->definition;
+    if ( !dstate->compound )
+    { dstate->compound = TRUE;
+      initSegStack(&dstate->stack, sizeof(Word),
+		   sizeof(dstate->buffer), dstate->buffer);
+    }
+    pushSegStack(&dstate->stack, dstate->term+1, Word);
+    //Sdprintf("Pushed %p\n", dstate->term+1);
+    dstate->term = &f->arguments[0];
+    return TRUE;
+  } else
+  { dstate->term++;
+
+    if ( isIndirect(*p) )
+    { *key = trie_intern_indirect(state->trie, *p, FALSE PASS_LD);
+      return *key != 0;
+    } else
+    { *key = *p;
+      return TRUE;
+    }
+  }
+}
+
+
 trie_choice *
-add_choice(trie_gen_state *state, trie_node *node)
+add_choice(trie_gen_state *state, descent_state *dstate, trie_node *node ARG_LD)
 { trie_children children = node->children;
-  trie_choice *ch = allocFromBuffer(&state->choicepoints, sizeof(*ch));
+  trie_choice *ch;
+  int has_key;
+  word k;
+
+  if ( dstate->prune )
+  { if ( !(has_key = get_key(state, dstate, &k PASS_LD)) )
+      dstate->prune = FALSE;
+  } else
+    has_key = FALSE;
 
   if ( children.any )
   { switch( children.any->type )
     { case TN_KEY:
+      if ( !has_key ||
+	   k == children.key->key ||
+	   tagex(children.key->key) == TAG_VAR ||
+	   IS_TRIE_KEY_POP(children.key->key) )
       {	word key   = children.key->key;
 
+	ch = allocFromBuffer(&state->choicepoints, sizeof(*ch));
 	ch->key    = key;
 	ch->child  = children.key->child;
         ch->choice.any = NULL;
-	break;
-      }
-      case TN_HASHED:
-      { void *k, *v;
 
-	ch->choice.table = newTableEnum(children.hash->table);
-        advanceTableEnum(ch->choice.table, &k, &v);
-	ch->key   = (word)k;
-	ch->child = (trie_node*)v;
+	if ( IS_TRIE_KEY_POP(children.key->key) && dstate->compound )
+	{ popSegStack(&dstate->stack, &dstate->term, Word);
+	  //Sdprintf("Popped %p\n", dstate->term);
+	}
+	break;
+      } else
+	return NULL;
+      case TN_HASHED:
+	if ( has_key && children.hash->var_keys == 0 )
+	{ trie_node *child;
+
+	  if ( (child = lookupHTable(children.hash->table, (void*)k)) )
+	  { ch = allocFromBuffer(&state->choicepoints, sizeof(*ch));
+	    ch->key = k;
+	    ch->child = child;
+	    ch->choice.any = NULL;
+	  } else
+	    return NULL;
+	} else
+	{ void *k, *v;
+
+	  dstate->prune = FALSE;
+	  ch = allocFromBuffer(&state->choicepoints, sizeof(*ch));
+	  ch->choice.table = newTableEnum(children.hash->table);
+	  advanceTableEnum(ch->choice.table, &k, &v);
+	  ch->key   = (word)k;
+	  ch->child = (trie_node*)v;
+	}
 	break;
-      }
       default:
 	assert(0);
     }
   } else
-  { memset(ch, 0, sizeof(*ch));
+  { ch = allocFromBuffer(&state->choicepoints, sizeof(*ch));
+    memset(ch, 0, sizeof(*ch));
     ch->child = node;
   }
 
@@ -1473,13 +1616,13 @@ add_choice(trie_gen_state *state, trie_node *node)
 }
 
 
-static int
-descent_node(trie_gen_state *state, trie_choice *ch)
-{ while( ch->child->children.any )
-  { ch = add_choice(state, ch->child);
+static trie_choice *
+descent_node(trie_gen_state *state, descent_state *dstate, trie_choice *ch ARG_LD)
+{ while( ch && ch->child->children.any )
+  { ch = add_choice(state, dstate, ch->child PASS_LD);
   }
 
-  return ch->child->value != 0;
+  return ch;
 }
 
 
@@ -1500,15 +1643,14 @@ advance_node(trie_choice *ch)
 }
 
 
-static int
-next_choice(trie_gen_state *state)
+static trie_choice *
+next_choice0(trie_gen_state *state, descent_state *dstate ARG_LD)
 { trie_choice *btm = base_choice(state);
   trie_choice  *ch = top_choice(state)-1;
 
   while(ch >= btm)
-  { if ( advance_node(ch) &&
-	 descent_node(state, ch) )
-      return TRUE;
+  { if ( advance_node(ch) )
+      return descent_node(state, dstate, ch PASS_LD);
 
     if ( ch->choice.table )
       freeTableEnum(ch->choice.table);
@@ -1521,6 +1663,22 @@ next_choice(trie_gen_state *state)
 }
 
 
+static int
+next_choice(trie_gen_state *state ARG_LD)
+{ trie_choice *ch;
+  descent_state dstate;
+
+  dstate.prune    = FALSE;
+  dstate.compound = FALSE;
+
+  do
+  { ch = next_choice0(state, &dstate PASS_LD);
+  } while (ch && ch->child->value == 0);
+
+  return ch != NULL;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Unify term with the term represented a trie path (list of trie_choice).
 Returns one of TRUE, FALSE or *_OVERFLOW.
@@ -1565,12 +1723,27 @@ trie_gen(term_t Trie, term_t Key, term_t Value,
 
       if ( get_trie(Trie, &trie) )
       { if ( trie->root.children.any )
-	{ acquire_trie(trie);
+	{ trie_choice *ch;
+	  descent_state dstate;
+	  int rc;
+
+	  TRIE_STAT_INC(trie, gen_call);
+
+	  dstate.term     = valTermRef(Key);
+	  dstate.compound = FALSE;
+	  dstate.prune	  = TRUE;
+	  deRef(dstate.term);
+
+	  acquire_trie(trie);
 	  state = &state_buf;
 	  init_trie_state(state, trie);
-	  if ( !descent_node(state, add_choice(state, &trie->root)) &&
-	       !next_choice(state) )
-	  { clear_trie_state(state);
+	  rc = ( (ch = add_choice(state, &dstate, &trie->root PASS_LD)) &&
+		 (ch = descent_node(state, &dstate, ch PASS_LD)) &&
+		 (ch->child->value || next_choice(state PASS_LD)) );
+	  clear_descent_state(&dstate);
+	  if ( !rc )
+	  { TRIE_STAT_INC(trie, gen_fail);
+	    clear_trie_state(state);
 	    return FALSE;
 	  }
 	  break;
@@ -1591,7 +1764,7 @@ trie_gen(term_t Trie, term_t Key, term_t Value,
   }
 
   fid = PL_open_foreign_frame();
-  for( ; !isEmptyBuffer(&state->choicepoints); next_choice(state) )
+  for( ; !isEmptyBuffer(&state->choicepoints); next_choice(state PASS_LD) )
   { int rc;
 
     for(;;)
@@ -1613,7 +1786,8 @@ trie_gen(term_t Trie, term_t Key, term_t Value,
 
     if ( (!Value || unify_value(Value, n->value PASS_LD)) &&
 	 (!Data  || unify_data(Data, n, ctx PASS_LD)) )
-    { if ( next_choice(state) )
+    { TRIE_STAT_INC(state->trie, gen_exit);
+      if ( next_choice(state PASS_LD) )
       { if ( !state->allocated )
 	{ trie_gen_state *nstate = allocForeignState(sizeof(*state));
 	  TmpBuffer nchp = &nstate->choicepoints;
@@ -1642,7 +1816,9 @@ trie_gen(term_t Trie, term_t Key, term_t Value,
 	return TRUE;
       }
     } else
+    { TRIE_STAT_INC(state->trie, gen_fail);
       PL_rewind_foreign_frame(fid);
+    }
 
 next:;
   }
@@ -1682,6 +1858,20 @@ PRED_IMPL("$trie_property", 2, trie_property, 0)
 { PRED_LD
   trie *trie;
 
+#ifdef O_TRIE_STATS
+  static atom_t ATOM_lookup_count = 0;
+  static atom_t ATOM_gen_call_count = 0;
+  static atom_t ATOM_gen_exit_count = 0;
+  static atom_t ATOM_gen_fail_count = 0;
+
+  if ( !ATOM_lookup_count )
+  { ATOM_lookup_count   = PL_new_atom("lookup_count");
+    ATOM_gen_call_count = PL_new_atom("gen_call_count");
+    ATOM_gen_exit_count = PL_new_atom("gen_exit_count");
+    ATOM_gen_fail_count = PL_new_atom("gen_fail_count");
+  }
+#endif
+
   if ( get_trie(A1, &trie) )
   { atom_t name; size_t arity;
 
@@ -1697,11 +1887,29 @@ PRED_IMPL("$trie_property", 2, trie_property, 0)
       } else if ( name == ATOM_size )
       { trie_stats stats;
 	stat_trie(trie, &stats);
+	if ( stats.nodes != trie->node_count )
+	  Sdprintf("OOPS: trie_property/2: counted %zd nodes, admin says %zd\n",
+		   stats.nodes, trie->node_count);
+	if ( stats.values != trie->value_count )
+	  Sdprintf("OOPS: trie_property/2: counted %zd values, admin says %zd\n",
+		   stats.values, trie->value_count);
+	// assert(stats.nodes  == trie->node_count);
+	// assert(stats.values == trie->value_count);
 	return PL_unify_int64(arg, stats.bytes);
       } else if ( name == ATOM_hashed )
       { trie_stats stats;
 	stat_trie(trie, &stats);
 	return PL_unify_int64(arg, stats.hashes);
+#ifdef O_TRIE_STATS
+      } else if ( name == ATOM_lookup_count )
+      { return PL_unify_int64(arg, trie->stats.lookups);
+      } else if ( name == ATOM_gen_call_count)
+      { return PL_unify_int64(arg, trie->stats.gen_call);
+      } else if ( name == ATOM_gen_exit_count)
+      { return PL_unify_int64(arg, trie->stats.gen_exit);
+      } else if ( name == ATOM_gen_fail_count)
+      { return PL_unify_int64(arg, trie->stats.gen_fail);
+#endif
       }
     }
   }
